# Copyright (C) 2022 Stefano Babic
#
# SPDX-License-Identifier: GPLv3

__all__ = [
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
]

__title__ = "SWUGenerator"
__summary__ = "SWU Package Generator for SWUpdate"
__uri__ = "https://github.com/sbabic/swugenerator"

__version__ = "0.6"

__author__ = "Stefano Babic"
__email__ = "stefano.babic@swupdate.org"

__license__ = "GPL version 3"
__copyright__ = f"Copyright 2022-2024 {__author__}"



# Copyright (C) 2025 INgo Rah
#
# SPDX-License-Identifier: GPLv3
import logging
from tempfile import TemporaryDirectory
from pathlib import Path

from swugenerator.swu_file import SWUFile

class SWUSigner:
    def __init__(self, in_file, out_file, crypt):
        self.signtool = crypt
        self.fout = open(out_file, "wb")
        self.cpio_out = SWUFile(self.fout)
        self.fin = open(in_file, "rb")
        self.cpio_in = SWUFile(self.fin)
        self.temp = TemporaryDirectory()

    def close(self):
        self.temp.cleanup()
        self.fin.close()
        self.cpio_out.add_trailer()
        self.fout.close()

    def process(self):
        # extract the swu file and get the file list
        artifacts = self.cpio_in.extract(self.temp.name)

        # call signing to generate the sig file
        if self.signtool:
            sw_desc_in = Path(self.temp.name) / artifacts[0]
            sw_desc_out = sw_desc_in.with_suffix(".sig")
            self.signtool.prepare_cmd(sw_desc_in, sw_desc_out)
            self.signtool.sign()
            if artifacts[0] + ".sig" not in artifacts:
                artifacts.insert(1, artifacts[0] + ".sig")
            else:
                logging.info("resigning a signed file")

        for name in artifacts:
            path = Path(self.temp.name) / name
            self.cpio_out.addartifacttoswu(path)




# pylint: disable=C0114,C0116,W0621
"""This file hosts integration tests to ensure tool creates valid SWUs"""
from pathlib import Path
import pytest
import shutil

import libarchive

from swugenerator import main

VALID_KEY = "390ad54490a4a5f53722291023c19e08ffb5c4677a59e958c96ffa6e641df040"
VALID_IV = "d5d601bacfe13100b149177318ebc7a4"
UPDATE_FILES = [
    "rootfs.ubifs",
    "swupdate.ext3.gz.u-boot",
    "sdcard.ext3.gz",
    "bootlogo.bmp",
    "uImage.bin",
    "fpga.txt",
    "bootloader-env",
    "README",
    "erase_at_end",
    "display_info",
]


@pytest.fixture()
def input_directory(tmp_path_factory):
    """Creates a directory to test"""
    return tmp_path_factory.mktemp("input_files")


@pytest.fixture()
def output_directory(tmp_path_factory):
    """Creates a directory to test"""
    return tmp_path_factory.mktemp("output_files")


@pytest.fixture()
def artifactory(input_directory):
    """Creates test files that match the filenames in the sw-description"""
    for filename in UPDATE_FILES:
        file = input_directory / filename
        with file.open("w") as file_fd:
            file_fd.write("THIS IS A TEST UPDATE FILE\n")
    return input_directory.resolve()


def copy_to_test_dir(filename, test_dir) -> Path:
    """Helper function to copy integration test inputs to test directory"""
    source_path = Path(__file__).parent / "test_data/integration_input" / filename
    dest_path = test_dir / filename
    shutil.copy(source_path, dest_path)
    return dest_path.resolve()


@pytest.fixture()
def sw_description_template(input_directory):
    """Copies sw-description template to test directory"""
    filename = "sw-description.in"
    return copy_to_test_dir(filename, input_directory)


@pytest.fixture()
def encrypted_sw_description_template(input_directory):
    """Copies sw-description template to test directory"""
    filename = "enc-sw-description.in"
    return copy_to_test_dir(filename, input_directory)


@pytest.fixture()
def config_file(input_directory):
    """Copies config file to test directory"""
    filename = "config"
    return copy_to_test_dir(filename, input_directory)


@pytest.fixture()
def signing_key(input_directory):
    """Creates signing key"""
    filename = "private.pem"
    return copy_to_test_dir(filename, input_directory)


@pytest.fixture()
def encryption_key(input_directory):
    """Creates encryption key"""
    encryption_key_filename = "valid_key.txt"
    encryption_key_path = input_directory / encryption_key_filename
    with encryption_key_path.open("w") as key_key_fd:
        key_key_fd.write(f"key={VALID_KEY}\niv={VALID_IV}")
    return encryption_key_path


def validate_swu(output_swu, encrypted=False):
    """Helper function to validate the output file against a pre-generated output file"""
    reference_swu = (
        Path(__file__).parent / "test_data/integration_output" / output_swu.name
    )
    with libarchive.Archive(output_swu.open()) as output, libarchive.Archive(
        reference_swu.open()
    ) as reference:
        for output_file, reference_file in zip(output, reference):
            if output_file.pathname != reference_file.pathname:
                print("Names do not match")
                return False
            if output_file.size != reference_file.size:
                print("Sizes do not match")
                return False
            if not encrypted and output.read(output_file.size) != reference.read(
                output_file.size
            ):
                print("File contents are not equal")
                return False
    return True


def test_basic_command_creates_valid_swu(
    artifactory, sw_description_template, config_file, output_directory
):
    output_file = "output.swu"
    command_args = [
        "-s",
        str(sw_description_template),
        "-a",
        str(artifactory),
        "-c",
        str(config_file),
        "-o",
        str((output_directory / output_file).resolve()),
        "create",
    ]
    main.parse_args(command_args)
    assert validate_swu(output_directory / output_file)


def test_command_with_sign_flag_creates_valid_signed_swu(
    artifactory, sw_description_template, config_file, signing_key, output_directory
):
    output_file = "signed_output.swu"
    command_args = [
        "-s",
        str(sw_description_template),
        "-a",
        str(artifactory),
        "-c",
        str(config_file),
        "-k",
        f"RSA,{signing_key}",
        "-o",
        str((output_directory / output_file).resolve()),
        "create",
    ]
    main.parse_args(command_args)
    assert validate_swu(output_directory / output_file)


def test_command_with_encryption_flag_creates_encrypted_swu(
    artifactory,
    encrypted_sw_description_template,
    config_file,
    encryption_key,
    output_directory,
):
    output_file = "output.swu.enc"
    command_args = [
        "-s",
        str(encrypted_sw_description_template),
        "-a",
        str(artifactory),
        "-c",
        str(config_file),
        "-K",
        str(encryption_key),
        "-o",
        str((output_directory / output_file).resolve()),
        "create",
    ]
    main.parse_args(command_args)
    assert validate_swu(output_directory / output_file, encrypted=True)


def test_command_with_sign_flag_and_encrypt_flag_creates_signed_encrypted_swu(
    artifactory,
    encrypted_sw_description_template,
    config_file,
    signing_key,
    encryption_key,
    output_directory,
):
    output_file = "signed_output.swu.enc"
    command_args = [
        "-s",
        str(encrypted_sw_description_template),
        "-a",
        str(artifactory),
        "-c",
        str(config_file),
        "-k",
        f"RSA,{signing_key}",
        "-K",
        str(encryption_key),
        "-o",
        str((output_directory / output_file).resolve()),
        "create",
    ]
    main.parse_args(command_args)
    assert validate_swu(output_directory / output_file, encrypted=True)

def test_creates_and_signs_valid_swu(
    artifactory, sw_description_template, config_file,
    signing_key, output_directory
):
    output_file = "output.swu"
    command_args = [
        "-s",
        str(sw_description_template),
        "-a",
        str(artifactory),
        "-c",
        str(config_file),
        "-o",
        str((output_directory / output_file).resolve()),
        "create",
    ]
    main.parse_args(command_args)
    signed_output_file = "signed_output.swu"
    command_args = [
        "-o",
        str((output_directory / signed_output_file).resolve()),
        "-k",
        f"RSA,{signing_key}",
        "sign",
        "-i",
        str((output_directory / output_file).resolve()),
    ]
    main.parse_args(command_args)

    assert validate_swu(output_directory / signed_output_file)


def test_with_encryption_flag_creates_encrypted_and_sign_swu(
    artifactory,
    encrypted_sw_description_template,
    config_file,
    encryption_key,
    signing_key,
    output_directory,
):
    output_file = "output.swu.enc"
    command_args = [
        "-s",
        str(encrypted_sw_description_template),
        "-a",
        str(artifactory),
        "-c",
        str(config_file),
        "-K",
        str(encryption_key),
        "-o",
        str((output_directory / output_file).resolve()),
        "create",
    ]
    main.parse_args(command_args)
    assert validate_swu(output_directory / output_file, encrypted=True)

    signed_output_file = "signed_output.swu.enc"
    command_args = [
        "-o",
        str((output_directory / signed_output_file).resolve()),
        "-k",
        f"RSA,{signing_key}",
        "sign",
        "-i",
        str((output_directory / output_file).resolve()),
    ]
    main.parse_args(command_args)

    assert validate_swu(output_directory / signed_output_file, encrypted=True)



# Copyright (C) 2022 Stefano Babic
#
# SPDX-License-Identifier: GPLv3
#
# This class manages to pack the SWU file and
# extracts a SWU file with returning the list.
# It has methods to add files to the archive.
import os
import stat


class CPIOException(Exception):
    pass

class MagicException(Exception):
    pass

class FormatException(Exception):
    pass

class SWUFile:
    def __init__(self, file):
        """
        :type file: FileIO of bytes
        """
        self.position = 0
        self.file = file
        self.renumbered_inode_count = 0
        self.artifacts = []
        # for reading
        self.header = None

    def _rawwrite(self, data):
        """
        :type data: bytes
        """
        self.file.write(data)
        self.position += len(data)

    def _align(self):
        offset = -(self.position % -4)
        if offset:
            self._rawwrite(b"\x00" * offset)

    def cpiocrc(self, cpio_filename):
        crc = 0
        with open(cpio_filename, "rb") as rd:
            chunk = rd.read(65536)
            while chunk:
                for b in chunk:
                    crc += b
                chunk = rd.read(65536)
        crc = crc & 0xFFFFFFFF
        return crc

    def addartifacttoswu(self, cpio_filename):
        """
        :type cpio_filename: string
        """

        statres = os.stat(cpio_filename)
        size = statres.st_size

        # very common error, so check sooner
        if size > 0xFFFFFFFF:
            raise CPIOException(
                "Too big file size for this CPIO format", statres.st_size
            )

        if not stat.S_ISREG(statres.st_mode) or not size:
            raise CPIOException("File is not a regular file")

        if cpio_filename == b"TRAILER!!!":
            raise CPIOException("Attempt to pass reserved filename", cpio_filename)

        self.write_header(cpio_filename)

        self._align()
        with open(cpio_filename, "rb") as xxx:
            while size:
                chunk = xxx.read(65536)
                if not chunk:
                    break
                self._rawwrite(chunk)
                size -= len(chunk)
        if size:
            raise CPIOException("File was changed while reading", cpio_filename)
        self.artifacts.append(cpio_filename)

    def next_renumbered_inode(self):
        """Make renumbered inode to be safe on 64 bit file systems.

        Simply counting up inodes is enough since swupdate doesn't preserve hard links.
        """
        self.renumbered_inode_count += 1
        return self.renumbered_inode_count

    def write_header(self, cpio_filename):
        if cpio_filename != "TRAILER!!!":
            statres = os.stat(cpio_filename)
            crc = self.cpiocrc(cpio_filename)
            base_filename = os.path.basename(cpio_filename)
            fields = [
                self.next_renumbered_inode(),
                statres.st_mode,
                statres.st_uid,
                statres.st_gid,
                statres.st_nlink,
                0,  # mtime
                statres.st_size,
                os.major(statres.st_dev),
                os.minor(statres.st_dev),
                os.major(statres.st_rdev),
                os.minor(statres.st_rdev),
                # Yes, length including last zero byte
                len(base_filename) + 1,
                # New CRC Format
                #  The ``CRC'' format is mis-named, as it uses a simple checksum and not a
                #  cyclic redundancy check.
                crc,
            ]
        else:
            base_filename = cpio_filename
            fields = [0, 0, 0, 0, 16, 0, 0, 0, 0, 0, 0, len(base_filename) + 1, 0]

        self._align()
        self._rawwrite(b"070702")
        for i, value in enumerate(fields):
            # UNIX epoch overflow, negative timestamps and so on...
            if (value > 0xFFFFFFFF) or (value < 0):
                raise CPIOException("STOP: value out of range", i, value)
            s = f"{value:08X}"
            self._rawwrite(bytes(s, "ascii"))

        fn = bytes(base_filename, "ascii") + b"\x00"
        self._rawwrite(fn)

    def add_trailer(self):
        try:
            self.write_header("TRAILER!!!")
            offset = -(self.position % -512)
            if offset:
                self._rawwrite(b"\x00" * offset)
        except PermissionError:
            # the file may have been opened for reading
            # flush shouldn't be called
            logging.info("flush called on read only file")

        self.position = 0

    def _extract_file(self, dir, name):
        c_mode = int(self.header[14:22], 16)
        c_filesize = int(self.header[54:62], 16)
        # Read file data
        filedata = self.file.read(c_filesize)
        pad = (4 - c_filesize % 4) % 4
        self.file.read(pad)
        # Compute output path
        path = os.path.join(dir, name)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        # check if its really a file
        if c_mode & 0o170000 != 0o040000:
            with open(path, "wb") as out:
                out.write(filedata)
            try:
                # preserve mode
                os.chmod(path, c_mode & 0o7777)
            except Exception:
                pass
        # check CRC
        outcrc = self.cpiocrc(path)
        incrc = int(self.header[102:110], 16)
        if outcrc !=  incrc:
            raise FormatException(
                f"Wrong file crc {incrc} vs. {outcrc}"
            )

    def _extract_next(self, dir):
        # Read fixed-size header (110 bytes for "newc" format)
        self.header = self.file.read(110)
        if len(self.header) < 110:
            raise FormatException(
                f"Unknown or unsupported header format. "
            )
        # Parse fields: all are ASCII hex numbers
        c_magic = self.header[0:6].decode()
        if c_magic != "070702":
            raise MagicException(
                f"Unknown or unsupported CPIO format. Our magic: {c_magic}"
            )
        # Parse fields (hex strings)
        c_namesize = int(self.header[94:102], 16)
        # Read filename (padded to 4 bytes)
        name = self.file.read(c_namesize)
        name = name[:-1].decode()
        if name == "TRAILER!!!":
            # end of archive
            return False
        # Align to 4 bytes
        pad = (4 - (110 + c_namesize) % 4) % 4
        self.file.read(pad)
        self.artifacts.append(name)
        self._extract_file(dir, name)

		# lets continue
        return True

    def extract(self, dir):
        # extract all files from the swu file
        next_file = True
        while next_file:
            next_file = self._extract_next(dir)

        return self.artifacts



# pylint: disable=C0114,C0116,W0621
import argparse
import libconf
import pytest

from swugenerator import main
from swugenerator.swu_sign import SWUSignCMS, SWUSignCustom, SWUSignPKCS11, SWUSignRSA

VALID_KEY = "390ad54490a4a5f53722291023c19e08ffb5c4677a59e958c96ffa6e641df040"
VALID_IV = "d5d601bacfe13100b149177318ebc7a4"
VALID_KEY_FILE = "valid_key.txt"
INVALID_KEY_FILE = "invalid_key.txt"


@pytest.fixture(scope="session")
def test_dir(tmp_path_factory):
    """Creates a directory to test"""
    test_space = tmp_path_factory.mktemp("archive")
    return test_space


#### Key file parsing tests ####
@pytest.fixture(scope="session")
def valid_key_file(test_dir):
    key_file = test_dir / VALID_KEY_FILE
    with key_file.open("w") as key_file_fd:
        key_file_fd.write(f"key={VALID_KEY}\niv={VALID_IV}")
    return key_file

@pytest.fixture(scope="session")
def valid_key_file_2(test_dir):
    key_file = test_dir / VALID_KEY_FILE
    with key_file.open("w") as key_file_fd:
        key_file_fd.write(f"key={VALID_KEY}\niv ={VALID_IV}")
    return key_file

@pytest.fixture(scope="session")
def invalid_key_file(test_dir):
    # Create invalid key file where only the key can be parsed
    key_file = test_dir / INVALID_KEY_FILE
    with key_file.open("w") as key_file_fd:
        key_file_fd.write(f"key foo\nkey={VALID_KEY}\nkey bar\niv\n{VALID_IV}")
    return key_file


def test_extract_keys_returns_valid_tuple_from_valid_file(valid_key_file):
    assert main.extract_keys(str(valid_key_file)) == (VALID_KEY, VALID_IV)

def test_extract_keys_returns_valid_tuple_from_valid_file_2(valid_key_file_2):
    assert main.extract_keys(str(valid_key_file_2)) == (VALID_KEY, VALID_IV)

def test_extract_keys_returns_none_from_key_file_thats_invalid(invalid_key_file):
    assert main.extract_keys(str(invalid_key_file)) == (VALID_KEY, None)


def test_extract_keys_returns_exception_from_key_file_that_dne():
    with pytest.raises(main.InvalidKeyFile):
        main.extract_keys("foo/bar/baz.txt")


#### Config file parsing tests ####
VALID_CONFIG = {"foo": 1, "bar": "test"}
VALID_CONFIG_FILE = "valid.cfg"
INVALID_CONFIG_FILE = "invalid.cfg"


@pytest.fixture(scope="session")
def valid_config_file(test_dir):
    config_file = test_dir / VALID_CONFIG_FILE
    with config_file.open("w") as config_file_fd:
        config_file_fd.write(libconf.dumps({"variables": VALID_CONFIG}))
    return config_file


@pytest.fixture(scope="session")
def invalid_config_file(test_dir):
    config_file = test_dir / VALID_CONFIG_FILE
    with config_file.open("w") as config_file_fd:
        config_file_fd.write("{" + libconf.dumps({"variables": VALID_CONFIG}))
    return config_file


def test_valid_config_file_is_properly_parsed(valid_config_file):
    assert main.parse_config_file(str(valid_config_file)) == VALID_CONFIG


def test_invalid_config_file_throws_exception(invalid_config_file):
    with pytest.raises(libconf.ConfigParseError):
        main.parse_config_file(str(invalid_config_file))


def test_missing_config_file_throws_exception():
    with pytest.raises(FileNotFoundError):
        main.parse_config_file("foo/bar/baz.txt")


#### Signing option parsing tests ####
SIGNING_TEST_PARAMETERS = [
    ("CMS,foo,bar,baz,qux", SWUSignCMS("foo", "bar", "baz", "qux")),
    ("CMS,foo,bar,,qux", SWUSignCMS("foo", "bar", "", "qux")),
    ("CMS,foo,bar,baz", SWUSignCMS("foo", "bar", "baz", None)),
    ("CMS,foo,bar", SWUSignCMS("foo", "bar", None, None)),
    ("RSA,foo,bar", SWUSignRSA("foo", "bar")),
    ("RSA,foo", SWUSignRSA("foo", None)),
    ("PKCS11,foo", SWUSignPKCS11("foo")),
    ("PKCS11,foo,bar", SWUSignPKCS11("foo", "bar")),
    ("CUSTOM,foo", SWUSignCustom("foo")),
]


@pytest.mark.parametrize("arg,expected", SIGNING_TEST_PARAMETERS)
def test_valid_siging_params_parsed_to_correct_signing_obj(arg, expected):
    signing_option = main.parse_signing_option(arg)
    assert type(signing_option) == type(expected)
    assert signing_option.type == expected.type
    assert signing_option.key == expected.key
    assert signing_option.cert == expected.cert
    assert signing_option.passin == expected.passin


INVALID_SIGNING_TEST_PARAMETERS = [
    ("CMS", "CMS requires private key, certificate, an optional password file and an optional file with additional certificates"),
    ("CMS,", "CMS requires private key, certificate, an optional password file and an optional file with additional certificates"),
    ("CMS,,", "CMS requires private key, certificate, an optional password file and an optional file with additional certificates"),
    ("CMS,,,", "CMS requires private key, certificate, an optional password file and an optional file with additional certificates"),
    ("CMS,,,,", "CMS requires private key, certificate, an optional password file and an optional file with additional certificates"),
    ("CMS,,,,,", "CMS requires private key, certificate, an optional password file and an optional file with additional certificates"),
    (
        "CMS,,foo,",
        "CMS requires private key, certificate, an optional password file and an optional file with additional certificates",
    ),
    ("CMS,foo", "CMS requires private key, certificate, an optional password file and an optional file with additional certificates"),
    (
        "CMS,foo,bar,baz,qux,jaz",
        "CMS requires private key, certificate, an optional password file and an optional file with additional certificates",
    ),
    ("RSA,foo,bar,baz", "RSA requires private key and an optional password file"),
    ("PKCS11", "PKCS11 requires pin and optional parameters such as module path, slot or id"),
    ("PKCS11,", "PKCS11 requires pin and optional parameters such as module path, slot or id"),
    ("PKCS11,,", "PKCS11 requires pin and optional parameters such as module path, slot or id"),
    ("CUSTOM", "CUSTOM requires custom command"),
    ("CUSTOM,", "CUSTOM requires custom command"),
    ("CUSTOM,,", "CUSTOM requires custom command"),
    ("CUSTOM,foo,", "CUSTOM requires custom command"),
    ("FOO", "Unknown signing command"),
    ("FOO,bar,baz", "Unknown signing command"),
]


@pytest.mark.parametrize("arg,exception_msg", INVALID_SIGNING_TEST_PARAMETERS)
def test_invalid_signing_params_throws_exception_with_correct_msg(arg, exception_msg):
    with pytest.raises(main.InvalidSigningOption) as error:
        main.parse_signing_option(arg)
    assert exception_msg in error.value.args


def test_parse_signing_option_cms_with_engine_and_keyform():
    # Only engine
    obj = main.parse_signing_option("CMS,foo,bar,baz,qux", engine="myengine")
    assert isinstance(obj, SWUSignCMS)
    assert obj.engine == "myengine"
    assert obj.key == "foo"
    assert obj.cert == "bar"
    assert obj.passin == "baz"
    assert obj.certfile == "qux"

    # Only keyform
    obj = main.parse_signing_option("CMS,foo,bar,baz,qux", keyform="mykeyform")
    assert isinstance(obj, SWUSignCMS)
    assert obj.keyform == "mykeyform"
    assert obj.key == "foo"
    assert obj.cert == "bar"
    assert obj.passin == "baz"
    assert obj.certfile == "qux"

    # Both engine and keyform
    obj = main.parse_signing_option("CMS,foo,bar,baz,qux", engine="myengine", keyform="mykeyform")
    assert isinstance(obj, SWUSignCMS)
    assert obj.engine == "myengine"
    assert obj.keyform == "mykeyform"
    assert obj.key == "foo"
    assert obj.cert == "bar"
    assert obj.passin == "baz"
    assert obj.certfile == "qux"

    # Neither engine nor keyform
    obj = main.parse_signing_option("CMS,foo,bar,baz,qux")
    assert isinstance(obj, SWUSignCMS)
    assert obj.engine is None
    assert obj.keyform is None
    assert obj.key == "foo"
    assert obj.cert == "bar"
    assert obj.passin == "baz"
    assert obj.certfile == "qux"


#### Parse args tests ####
@pytest.fixture
def mock_main_funcs(monkeypatch):
    def mock_create_swu(*_):
        return True

    def mock_extract_keys(*_):
        return "foo", "bar"

    def mock_parse_signing_option(*_):
        return SWUSignCMS("foo", "bar", "baz", "qux")

    def mock_parse_config_file(*_):
        return {}

    def mock_argparse_error(*_):
        raise Exception

    def mock_sign_swu(*_):
        return True

    monkeypatch.setattr(main, "create_swu", mock_create_swu)
    monkeypatch.setattr(main, "sign_swu", mock_sign_swu)
    monkeypatch.setattr(main, "extract_keys", mock_extract_keys)
    monkeypatch.setattr(main, "parse_signing_option", mock_parse_signing_option)
    monkeypatch.setattr(main, "parse_config_file", mock_parse_config_file)
    monkeypatch.setattr(argparse.ArgumentParser, "exit", mock_argparse_error)


VALID_COMMANDS = [
    (["-s", "sw-description", "-o", "test.swu", "create"]),
    (["--sw-description", "sw-description", "--swu-file", "test.swu", "create"]),
    (
        [
            "-K",
            "key.txt",
            "-n",
            "-e",
            "-x",
            "-k",
            "CUSTOM,foo",
            "-s",
            "sw-description",
            "-t",
            "-a",
            ".,..",
            "-o",
            "test.swu",
            "-c",
            "test.cfg",
            "-l",
            "DEBUG",
            "create",
        ]
    ),
    (
        [
            "--encryption-key-file",
            "key.txt",
            "--no-compress",
            "--no-encrypt",
            "--no-ivt",
            "--sign",
            "CUSTOM,foo",
            "--sw-description",
            "sw-description",
            "--encrypt-swdesc",
            "--artifactory",
            ".,..",
            "--swu-file",
            "test.swu",
            "--config",
            "test.cfg",
            "--loglevel",
            "DEBUG",
            "create",
        ]
    ),
    (["-o", "test.swu", "sign", "-i", "in.swu"]),
]


@pytest.mark.parametrize("args", VALID_COMMANDS)
def test_parsing_valid_args_doesnt_throw(args, mock_main_funcs):
    main.parse_args(args)


INVALID_COMMANDS = [
    (["create"]),
    (["-s", "sw-description", "-o", "test.swu"]),
    (["-s", "-o", "test.swu", "create"]),
    (["-s", "sw-description", "-o", "create"]),
    (["-s", "-o", "create"]),
    (["-s", "-o"]),
    (["-K", "-s", "sw-description", "-o", "test.swu", "create"]),
    (["-k", "-s", "sw-description", "-o", "test.swu", "create"]),
    (["-a", "sw-description", "-o", "test.swu", "create"]),
    (["-c", "sw-description", "-o", "test.swu", "create"]),
    (["-l", "sw-description", "-o", "test.swu", "create"]),
    (["-l", "baz", "sw-description", "-o", "test.swu", "create"]),
    (["sign"]),
    (["-o", "test.swu", "sign"]),
    (["-i", "-o", "test.swu", "sign"]),
    (["-i", "in.swu", "-o", "test.swu", "sign"]),
    (["-o", "test.swu", "create"]),
    (["-o", "sign", "-i", "in.swu"]),
]


@pytest.mark.parametrize("args", INVALID_COMMANDS)
def test_parsing_invalid_args_does_throw_argparse_exception(args, mock_main_funcs):
    with pytest.raises(Exception):
        main.parse_args(args)



# SWUGenerator, SWU Package Generator for SWUpdate
#
# Copyright (C) 2022 Stefano Babic
#
# SPDX-License-Identifier: GPLv3
# pylint: disable=C0114

import argparse
import logging
import os
import sys
import textwrap
from pathlib import Path
from typing import List, Optional, Tuple, Union

import libconf

from swugenerator import __about__, generator, signer

from swugenerator.swu_sign import SWUSignCMS, SWUSignCustom, SWUSignPKCS11, SWUSignRSA


class InvalidKeyFile(ValueError):
    """Raised when a key file is invalid"""


class InvalidSigningOption(ValueError):
    """Raised when an invalid signing option is passed via command line"""


def extract_keys(keyfile: str) -> Tuple[Optional[str], Optional[str]]:
    """Extracts encryption key and initialization vector (IV)

    TODO Ensure alignment of key file with description in docs:
    (https://sbabic.github.io/swupdate/encrypted_images.html#building-an-encrypted-swu-image)

    Args:
        keyfile (str): Path to file containing key and IV

    Raises:
        InvalidKeyFile: If key file cannot be read

    Returns:
        Tuple[Optional[str], Optional[str]]: Key and IV if found.
            Tuple will contain None for either value if missing
    """
    try:
        with open(keyfile, "r", encoding="utf-8") as keyfile_fd:
            lines = keyfile_fd.readlines()
    except IOError as error:
        raise InvalidKeyFile(f"Failed to open key file {keyfile}") from error

    enc_key, init_vec = None, None
    for line in lines:
        key, value = (
            line.replace(" ", "").rstrip("\n").split("=") if len(line.split("=")) == 2 else (None, None)
        )
        if key == "key":
            enc_key = value
        if key == "iv":
            init_vec = value
    return enc_key, init_vec


def parse_config_file(config_file_arg: str) -> dict:
    """Parses configuration file to store key value pairs

    Args:
        config_file_arg (str): Path to config file

    Returns:
        dict: Key value pairs parsed from configuration file
    """
    config_vars = {}
    with open(config_file_arg, "r", encoding="utf-8") as config_fd:
        config = libconf.load(config_fd)
        for key, keydict in config.items():
            if key == "variables":
                for varname, varvalue in keydict.items():
                    logging.debug("VAR = %s VAL = %s", varname, varvalue)
                    config_vars[varname] = varvalue
    return config_vars


def parse_signing_option(
    sign_arg: str,
    engine: str = None,
    keyform: str = None,
) -> Union[SWUSignCMS, SWUSignRSA, SWUSignPKCS11, SWUSignCustom]:
    """Parses signgning option passed by user. Valid options can be found below.

    CMS,<private key>,<certificate used to sign>,<file with password>,<file with certs>
    CMS,<private key>,<certificate used to sign>,<file with password>
    CMS,<private key>,<certificate used to sign>
    RSA,<private key>,<file with password>
    RSA,<private key>
    PKCS11,<pin>[,<module>,<slot>,<id>]
    CUSTOM,<custom command>

    Args:
        sign_arg (str): argument passed by user
        engine (str): OpenSSL engine to use for signing (e.g., pkcs11)
        keyform (str): Key format to use for signing (e.g., engine)

    Raises:
        InvalidSigningOption: If option passed by user is invalid

    Returns:
        Union[SWUSignCMS, SWUSignRSA, SWUSignPKCS11, SWUSignCustom]: Signing option to use
    """
    sign_parms = sign_arg.split(",")
    cmd = sign_parms[0]
    if cmd == "CMS":
        if len(sign_parms) not in (3, 4, 5) or not all(sign_parms[0:2]):
            raise InvalidSigningOption(
                "CMS requires private key, certificate, an optional password file and an optional file with additional certificates"
            )
        # Format : CMS,<private key>,<certificate used to sign>,<file with password>,<file with certs>
        if len(sign_parms) == 5:
            return SWUSignCMS(sign_parms[1], sign_parms[2], sign_parms[3], sign_parms[4], engine, keyform)
        # Format : CMS,<private key>,<certificate used to sign>,<file with password>
        elif len(sign_parms) == 4:
            return SWUSignCMS(sign_parms[1], sign_parms[2], sign_parms[3], None, engine, keyform)
        # Format : CMS,<private key>,<certificate used to sign>
        else:
            return SWUSignCMS(sign_parms[1], sign_parms[2], None, None, engine, keyform)
    if cmd == "RSA":
        if len(sign_parms) not in (2, 3) or not all(sign_parms):
            raise InvalidSigningOption(
                "RSA requires private key and an optional password file"
            )
        # Format : RSA,<private key>,<file with password>
        if len(sign_parms) == 3:
            return SWUSignRSA(sign_parms[1], sign_parms[2])
        # Format : RSA,<private key>
        return SWUSignRSA(sign_parms[1], None)
    if cmd == "PKCS11":
        # Format : PKCS11,<pin>[,<module>,<slot>,<id>]
        if len(sign_parms) not in range(2, 6) or not all(sign_parms[0:2]):
            raise InvalidSigningOption("PKCS11 requires pin and optional parameters such as module path, slot or id")
        pin = sign_parms[1]
        module = sign_parms[2] if len(sign_parms) > 2 else None
        slot   = sign_parms[3] if len(sign_parms) > 3 else None
        obj_id = sign_parms[4] if len(sign_parms) > 4 else None

        return SWUSignPKCS11(pin, module, slot, obj_id)
    if cmd == "CUSTOM":
        # Format : CUSTOM,<custom command>
        if len(sign_parms) < 2 or not all(sign_parms):
            raise InvalidSigningOption("CUSTOM requires custom command")
        return SWUSignCustom(sign_parms[1:])
    raise InvalidSigningOption("Unknown signing command")


def set_log_level(arg: str) -> str:
    """Sets log level

    This is meant to be used with Argparse's type param for
    add_argument, but this allows log level to be set when
    argparse parses commands.

    Args:
        arg (str): Log level in string form (All caps)

    Returns:
        str: Returns arg to be parsed by argparse.
            Argparse will then make sure arg is in choices
            and return an error to the user otherwise.
    """
    if arg in ("DEBUG", "INFO", "ERROR", "CRITICAL"):
        logging.basicConfig(level=logging.getLevelName(arg))
    return arg


def create_swu(args: argparse.Namespace) -> None:
    """Creates SWU archive from arguments passed to SWUGenerate

    Args:
        args (argparse.Namespace): Parsed arguments to generate SWU file with
    """
    # Extract key and iv from encryption_key_file (Will default to '(None, None)')
    key, init_vec = args.encryption_key_file

    # Add current working directory to search path
    args.artifactory.append(Path(os.getcwd()))

    # Pass engine and keyform to parse_signing_option if sign is set
    if hasattr(args, 'sign') and args.sign and isinstance(args.sign, str):
        args.sign = parse_signing_option(args.sign, args.engine, args.keyform)

    swu = generator.SWUGenerator(
        args.sw_description,
        args.swu_file,
        args.config,
        args.artifactory,
        args.sign,
        key,
        init_vec,
        args.encrypt_swdesc,
        args.no_compress,
        args.no_encrypt,
        args.no_ivt,
        args.no_hash,
    )
    swu.process()
    swu.close()

def sign_swu(args: argparse.Namespace) -> None:
    swu = signer.SWUSigner(
        args.swu_in_file,
        args.swu_file,
        args.sign
    )
    swu.process()
    swu.close()

def parse_args(args: List[str]) -> None:
    """Sets up arguments for swugenerator and parses commandline args

    Args:
        args (List[str]): Command line arguments
    """
    parser = argparse.ArgumentParser(
        prog=__about__.__title__,
        description=__about__.__summary__ + " " + __about__.__version__,
        fromfile_prefix_chars="@",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-K",
        "--encryption-key-file",
        default=(None, None),
        type=extract_keys,
        help="<key,iv> : AES Key to encrypt artifacts",
    )

    parser.add_argument(
        "-n",
        "--no-compress",
        action="store_true",
        help="Do not compress files",
    )

    parser.add_argument(
        "-e",
        "--no-encrypt",
        action="store_true",
        help="Do not encrypt files",
    )

    parser.add_argument(
        "-x",
        "--no-ivt",
        action="store_true",
        help="Do not generate IV when encrypting",
    )

    parser.add_argument(
        "-y",
        "--no-hash",
        action="store_true",
        help="Do not store sha256 hash in sw-description",
    )

    parser.add_argument(
        "-k",
        "--sign",
        help=textwrap.dedent(
            """\
            RSA key or certificate to sign the SWU
            One of :
            CMS,<private key>,<certificate used to sign>,<file with password if any>,<file with certs if any>
                 -g, --engine ENGINE       OpenSSL engine to use for signing (e.g., pkcs11)
                 -f, --keyform KEYFORM     Key format to use for signing (e.g., engine)
            RSA,<private key>,<file with password if any>
            PKCS11,<pin>[,<module>,<slot>,<id>]
            CUSTOM,<custom command> """
        ),
    )

    parser.add_argument(
        "-s",
        "--sw-description",
        required=False,
        type=lambda p: Path(p).resolve(),
        help="sw-description template, required for the create command",
    )

    parser.add_argument(
        "-t",
        "--encrypt-swdesc",
        action="store_const",
        const=True,
        default=False,
        help="Encrypt sw-description",
    )

    parser.add_argument(
        "-a",
        "--artifactory",
        default=[],
        type=lambda paths: [Path(p).resolve() for p in paths.split(",")],
        help="list of directories where artifacts are searched",
    )

    parser.add_argument(
        "-o",
        "--swu-file",
        required=True,
        type=Path,
        help="SWU output file",
    )

    parser.add_argument(
        "-c",
        "--config",
        default={},
        type=parse_config_file,
        help="configuration file",
    )

    parser.add_argument(
        "-l",
        "--loglevel",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        type=set_log_level,
        help="set log level, default is WARNING",
    )

    parser.add_argument(
        "-g",
        "--engine",
        default=None,
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-f",
        "--keyform",
        default=None,
        help=argparse.SUPPRESS,
    )

    subparsers = parser.add_subparsers(
        title="command", help="command to be executed", required=True
    )
    create_subparser = subparsers.add_parser("create", help="creates a SWU file")
    create_subparser.set_defaults(func=create_swu)
    sign_subparser = subparsers.add_parser("sign", help="signs an existing SWU file provided by -i")
    sign_subparser.set_defaults(func=sign_swu)
    sign_subparser.add_argument(
        "-i",
        "--swu-in-file",
        required=True,
        type=Path,
        help="SWU input file to be signed, required for the sign command",
    )
    args = parser.parse_args(args)
    if hasattr(args, 'sign') and args.sign:
        args.sign = parse_signing_option(args.sign, args.engine, args.keyform)
    if args.func == create_swu and not args.sw_description:
        parser.error(
            "the following arguments are required: -s/--sw-description"
        )

    args.func(args)


def main():
    """Main entry point for SWUGenerator"""
    # Arg parsing is in separate function
    # to allow argument parsing to be easily
    # tested with pytest
    parse_args(sys.argv[1:])


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Stefano Babic
#
# SPDX-License-Identifier: GPLv3
import copy
import logging
import subprocess
import sys


class SWUSign:
    def __init__(self):
        self.type = None
        self.key = None
        self.cert = None
        self.cmd = None
        self.passin = None
        self.certfile = None
        self.signcmd = []

    def get_passwd_file_args(self):
        passwd_args = []
        if self.passin:
            passwd_args = ["-passin", f"file:{self.passin}"]
        return passwd_args

    def set_password_file(self, passin):
        self.passin = passin

    def get_certfile_args(self):
        certfile_args = []
        if self.certfile:
            certfile_args = ["-certfile", self.certfile]
        return certfile_args

    def set_certfile(self, certfile):
        self.certfile = certfile

    def sign(self):
        try:
            subprocess.run(self.signcmd, check=True, text=True)
        except subprocess.CalledProcessError:
            logging.critical(
                "SWU cannot be signed, signing command was %s", self.signcmd
            )
            sys.exit(1)


class SWUSignCMS(SWUSign):
    def __init__(self, key, cert, passin, certfile, engine=None, keyform=None):
        super().__init__()
        self.type = "CMS"
        self.key = key
        self.cert = cert
        self.passin = passin
        self.certfile = certfile
        self.engine = engine
        self.keyform = keyform

    def prepare_cmd(self, sw_desc_in, sw_desc_sig):
        self.signcmd = [
            "openssl",
            "cms",
            "-sign",
            "-in",
            sw_desc_in,
            "-out",
            sw_desc_sig,
            "-signer",
            self.cert,
        ]
        if self.engine:
            self.signcmd += ["-engine", self.engine]
        if self.keyform:
            self.signcmd += ["-keyform", self.keyform]
        self.signcmd += [
            "-inkey",
            self.key,
            "-outform",
            "DER",
            "-nosmimecap",
            "-binary",
        ]
        self.signcmd += self.get_passwd_file_args()
        self.signcmd += self.get_certfile_args()


class SWUSignRSA(SWUSign):
    def __init__(self, key, passin):
        super().__init__()
        self.type = "RSA"
        self.key = key
        self.passin = passin

    def prepare_cmd(self, sw_desc_in, sw_desc_sig):
        self.signcmd = (
            ["openssl", "dgst", "-sha256", "-sign", self.key]
            + self.get_passwd_file_args()
            + ["-out", sw_desc_sig, sw_desc_in]
        )


class SWUSignCustom(SWUSign):
    def __init__(self, cmd):
        super().__init__()
        self.type = "CUSTOM"
        self.custom = cmd

    def prepare_cmd(self, sw_desc_in, sw_desc_sign):
        self.signcmd = self.custom
        self.signcmd.append(sw_desc_in)
        self.signcmd.append(sw_desc_sign)


# Note: tested with Nitrokey HSM
class SWUSignPKCS11(SWUSign):
    def __init__(self, pin, module=None, slot=None, obj_id=None):
        super().__init__()
        self.type = "PKCS11"
        self.custom = []
        if module:
            self.custom.extend(["--module", module])
        if slot:
            self.custom.extend(["--slot", slot])
        if obj_id:
            self.custom.extend(["--id", obj_id])
        self.custom.extend(["--pin", pin])

    def prepare_cmd(self, sw_desc_in, sw_desc_sig):
        self.signcmd = [
            "pkcs11-tool",
            "-s",
            "-m",
            "SHA256-RSA-PKCS",
            "-i",
            sw_desc_in,
            "-o",
            sw_desc_sig,
        ] + self.custom
